import { isNumber } from "lodash"
import {
  optionsHaveIgnored,
  ruleMessages,
  styleSearch,
  report,
  validateOptions,
} from "../../utils"

export const ruleName = "max-line-length"

export const messages = ruleMessages(ruleName, {
  expected: l => `Expected line length equal to or less than ${l} characters`,
})

export default function (maxLength, options) {
  return (root, result) => {
    const validOptions = validateOptions(result, ruleName, {
      maxLength: isNumber,
    }, {
      actual: options,
      possible: {
        ignore: ["non-comments"],
      },
      optional: true,
    })
    if (!validOptions) { return }

    // Collapse all urls into something nice and short,
    // so they do not throw the game
    const rootString = root.source.input.css.replace(/url\(.*\)/g, "url()")

    // Check first line
    checkNewline({ endIndex: 0 })

    // Check subsequent lines
    styleSearch({ source: rootString, target: ["\n"], checkComments: true }, checkNewline)

    function checkNewline(match) {
      let nextNewlineIndex = rootString.indexOf("\n", match.endIndex)

      // Accommodate last line
      if (nextNewlineIndex === -1) {
        nextNewlineIndex = rootString.length
      }

      // If the line's length is less than or equal to the specified
      // max, ignore it
      if (nextNewlineIndex - match.endIndex <= maxLength) { return }

      // If there are no spaces besides initial (indent) spaces, ignore it
      const lineString = rootString.slice(match.endIndex, nextNewlineIndex)
      if (lineString.replace(/^\s+/, "").indexOf(" ") === -1) {
        return
      }

      if (optionsHaveIgnored(options, "non-comments")) {
        // This trimming business is to notice when the line starts a
        // comment but that comment is indented, e.g.
        //       /* something here */
        const nextTwoChars = rootString.slice(match.endIndex).trim().slice(0, 2)
        if (
          !match.insideComment
          && (nextTwoChars !== "/*" && nextTwoChars !== "//")
        ) { return }
      }

      report({
        message: messages.expected(maxLength),
        node: root,
        index: nextNewlineIndex - 1,
        result,
        ruleName,
      })
    }
  }
}
